@@ -73,22 +73,12 @@ typedef struct
   GHashTable *status_icons;
   /* List of owned GCancellables for each pending avatar load operation */
   GList *avatar_cancellables;
+  /* Hash: FolksIndividual* -> GQueue (GtkTreeRowReference) */
+  GHashTable                  *folks_individual_cache;
+  /* Hash: char *groupname -> GtkTreeRowReference *row */
+  GHashTable                  *empathy_group_cache;
 } EmpathyIndividualStorePriv;
 
-typedef struct
-{
-  GtkTreeIter iter;
-  const gchar *name;
-  gboolean found;
-} FindGroup;
-
-typedef struct
-{
-  FolksIndividual *individual;
-  gboolean found;
-  GList *iters;
-} FindContact;
-
 typedef struct
 {
   EmpathyIndividualStore *self;
@@ -133,21 +123,25 @@ individual_can_audio_video_call (FolksIndividual *individual,
     {
       FolksPersona *persona = gee_iterator_get (iter);
       TpContact *tp_contact;
-      EmpathyContact *contact;
 
       if (!empathy_folks_persona_is_interesting (persona))
         goto while_finish;
 
       tp_contact = tpf_persona_get_contact (TPF_PERSONA (persona));
-      contact = empathy_contact_dup_from_tp_contact (tp_contact);
-      empathy_contact_set_persona (contact, persona);
+      if (tp_contact != NULL)
+        {
+          EmpathyContact *contact;
 
-      can_audio = can_audio || empathy_contact_get_capabilities (contact) &
-          EMPATHY_CAPABILITIES_AUDIO;
-      can_video = can_video || empathy_contact_get_capabilities (contact) &
-          EMPATHY_CAPABILITIES_VIDEO;
+          contact = empathy_contact_dup_from_tp_contact (tp_contact);
+          empathy_contact_set_persona (contact, persona);
 
-      g_object_unref (contact);
+          can_audio = can_audio || empathy_contact_get_capabilities (contact) &
+              EMPATHY_CAPABILITIES_AUDIO;
+          can_video = can_video || empathy_contact_get_capabilities (contact) &
+              EMPATHY_CAPABILITIES_VIDEO;
+
+          g_object_unref (contact);
+        }
 while_finish:
       g_clear_object (&persona);
 
@@ -190,7 +184,8 @@ individual_get_client_types (FolksIndividual *individual)
           presence_type = folks_presence_details_get_presence_type (presence);
 
           tp_contact = tpf_persona_get_contact (TPF_PERSONA (persona));
-          types = tp_contact_get_client_types (tp_contact);
+          if (tp_contact != NULL)
+            types = tp_contact_get_client_types (tp_contact);
         }
 
 while_finish:
@@ -207,8 +202,12 @@ add_individual_to_store (GtkTreeStore *self,
     GtkTreeIter *parent,
     FolksIndividual *individual)
 {
+  EmpathyIndividualStorePriv *priv = GET_PRIV (self);
   gboolean can_audio_call, can_video_call;
   const gchar * const *types;
+  GtkTreeRowReference *row_ref;
+  GtkTreePath *path;
+  GQueue *queue;
 
   individual_can_audio_video_call (individual, &can_audio_call,
       &can_video_call);
@@ -225,34 +224,22 @@ add_individual_to_store (GtkTreeStore *self,
       EMPATHY_INDIVIDUAL_STORE_COL_CAN_VIDEO_CALL, can_video_call,
       EMPATHY_INDIVIDUAL_STORE_COL_CLIENT_TYPES, types,
       -1);
-}
 
-static gboolean
-individual_store_get_group_foreach (GtkTreeModel *model,
-    GtkTreePath *path,
-    GtkTreeIter *iter,
-    FindGroup *fg)
-{
-  gchar *str;
-  gboolean is_group;
-
-  /* Groups are only at the top level. */
-  if (gtk_tree_path_get_depth (path) != 1)
-    return FALSE;
-
-  gtk_tree_model_get (model, iter,
-      EMPATHY_INDIVIDUAL_STORE_COL_NAME, &str,
-      EMPATHY_INDIVIDUAL_STORE_COL_IS_GROUP, &is_group, -1);
-
-  if (is_group && !tp_strdiff (str, fg->name))
+  path = gtk_tree_model_get_path (GTK_TREE_MODEL (self), iter);
+  row_ref = gtk_tree_row_reference_new (GTK_TREE_MODEL (self), path);
+  queue = g_hash_table_lookup (priv->folks_individual_cache, individual);
+  if (queue)
     {
-      fg->found = TRUE;
-      fg->iter = *iter;
+      g_queue_push_tail (queue, row_ref);
     }
-
-  g_free (str);
-
-  return fg->found;
+  else
+    {
+      queue = g_queue_new ();
+      g_queue_push_tail (queue, row_ref);
+      g_hash_table_insert (priv->folks_individual_cache, individual,
+          queue);
+    }
+  gtk_tree_path_free (path);
 }
 
 static void
@@ -263,21 +250,19 @@ individual_store_get_group (EmpathyIndividualStore *self,
     gboolean *created,
     gboolean is_fake_group)
 {
+  EmpathyIndividualStorePriv *priv = GET_PRIV (self);
   GtkTreeModel *model;
   GtkTreeIter iter_group;
   GtkTreeIter iter_separator;
-  FindGroup fg;
-
-  memset (&fg, 0, sizeof (fg));
-
-  fg.name = name;
+  GtkTreeRowReference         *row_ref;
 
   model = GTK_TREE_MODEL (self);
-  gtk_tree_model_foreach (model,
-      (GtkTreeModelForeachFunc) individual_store_get_group_foreach, &fg);
+  row_ref = g_hash_table_lookup (priv->empathy_group_cache, name);
 
-  if (!fg.found)
+  if (row_ref == NULL)
     {
+      GtkTreePath *path;
+
       if (created)
         *created = TRUE;
 
@@ -291,6 +276,12 @@ individual_store_get_group (EmpathyIndividualStore *self,
           EMPATHY_INDIVIDUAL_STORE_COL_IS_FAKE_GROUP, is_fake_group,
           -1);
 
+      path = gtk_tree_model_get_path (GTK_TREE_MODEL (self), &iter_group);
+      row_ref = gtk_tree_row_reference_new (GTK_TREE_MODEL (self), path);
+      g_hash_table_insert (priv->empathy_group_cache,
+      	g_strdup (name), row_ref);
+      gtk_tree_path_free (path);
+
       if (iter_group_to_set)
         *iter_group_to_set = iter_group;
 
@@ -304,13 +295,22 @@ individual_store_get_group (EmpathyIndividualStore *self,
     }
   else
     {
+      GtkTreePath *path = gtk_tree_row_reference_get_path (row_ref);
+      GtkTreeIter iter;
+
+      if (!gtk_tree_model_get_iter (model, &iter, path)) {
+      	gtk_tree_path_free (path);
+      	return;
+      }
+      gtk_tree_path_free (path);
+
       if (created)
         *created = FALSE;
 
       if (iter_group_to_set)
-        *iter_group_to_set = fg.iter;
+        *iter_group_to_set = iter;
 
-      iter_separator = fg.iter;
+      iter_separator = iter;
 
       if (gtk_tree_model_iter_next (model, &iter_separator))
         {
@@ -325,48 +325,37 @@ individual_store_get_group (EmpathyIndividualStore *self,
     }
 }
 
-static gboolean
-individual_store_find_contact_foreach (GtkTreeModel *model,
-    GtkTreePath *path,
-    GtkTreeIter *iter,
-    FindContact *fc)
-{
-  FolksIndividual *individual;
-
-  gtk_tree_model_get (model, iter,
-      EMPATHY_INDIVIDUAL_STORE_COL_INDIVIDUAL, &individual, -1);
-
-  if (individual == fc->individual)
-    {
-      fc->found = TRUE;
-      fc->iters = g_list_append (fc->iters, gtk_tree_iter_copy (iter));
-    }
-
-  tp_clear_object (&individual);
-
-  return FALSE;
-}
-
 static GList *
 individual_store_find_contact (EmpathyIndividualStore *self,
     FolksIndividual *individual)
 {
+  EmpathyIndividualStorePriv *priv = GET_PRIV (self);
   GtkTreeModel *model;
-  GList *l = NULL;
-  FindContact fc;
-
-  memset (&fc, 0, sizeof (fc));
-
-  fc.individual = individual;
+  GQueue *row_refs_queue;
+  GList *i;
+  GList *iters_list = NULL;
 
   model = GTK_TREE_MODEL (self);
-  gtk_tree_model_foreach (model,
-      (GtkTreeModelForeachFunc) individual_store_find_contact_foreach, &fc);
+  row_refs_queue = g_hash_table_lookup (priv->folks_individual_cache,
+      individual);
+  if (!row_refs_queue)
+    return NULL;
 
-  if (fc.found)
-    l = fc.iters;
+  for (i = g_queue_peek_head_link (row_refs_queue) ; i != NULL ; i = i->next)
+    {
+      GtkTreePath *path = gtk_tree_row_reference_get_path (i->data);
+      GtkTreeIter iter;
+      if (!gtk_tree_model_get_iter (model, &iter, path))
+        {
+          gtk_tree_path_free (path);
+          continue;
+        }
+      gtk_tree_path_free (path);
+      iters_list = g_list_prepend
+          (iters_list, gtk_tree_iter_copy (&iter));
+    }
 
-  return l;
+  return iters_list;
 }
 
 static void
@@ -380,36 +369,54 @@ static void
 individual_store_remove_individual (EmpathyIndividualStore *self,
     FolksIndividual *individual)
 {
+  EmpathyIndividualStorePriv *priv = GET_PRIV (self);
   GtkTreeModel *model;
-  GList *iters, *l;
+  GQueue *row_refs;
+  GList *l;
 
-  iters = individual_store_find_contact (self, individual);
-  if (iters == NULL)
+  row_refs = g_hash_table_lookup (priv->folks_individual_cache, individual);
+  if (!row_refs)
     return;
 
   /* Clean up model */
   model = GTK_TREE_MODEL (self);
 
-  for (l = iters; l; l = l->next)
+  for (l = g_queue_peek_head_link (row_refs); l; l = l->next)
     {
+      GtkTreePath *path = gtk_tree_row_reference_get_path (l->data);
+      GtkTreeIter iter;
       GtkTreeIter parent;
 
+      if (!gtk_tree_model_get_iter (GTK_TREE_MODEL (self), &iter,
+                                    path))
+        {
+          gtk_tree_path_free (path);
+          continue;
+        }
+      gtk_tree_path_free (path);
+
       /* NOTE: it is only <= 2 here because we have
        * separators after the group name, otherwise it
        * should be 1.
        */
-      if (gtk_tree_model_iter_parent (model, &parent, l->data) &&
+      if (gtk_tree_model_iter_parent (model, &parent, &iter) &&
           gtk_tree_model_iter_n_children (model, &parent) <= 2)
         {
+          gchar *group_name;
+          gtk_tree_model_get (model, &parent,
+              EMPATHY_CONTACT_LIST_STORE_COL_NAME, &group_name,
+              -1);
+          g_hash_table_remove (priv->empathy_group_cache,
+              group_name);
           gtk_tree_store_remove (GTK_TREE_STORE (self), &parent);
         }
       else
         {
-          gtk_tree_store_remove (GTK_TREE_STORE (self), l->data);
+          gtk_tree_store_remove (GTK_TREE_STORE (self), &iter);
         }
     }
 
-  free_iters (iters);
+  g_hash_table_remove (priv->folks_individual_cache, individual);
 }
 
 static void
@@ -900,14 +907,17 @@ individual_personas_changed_cb (FolksIndividual *individual,
       if (TPF_IS_PERSONA (persona))
         {
           tp_contact = tpf_persona_get_contact (persona);
-          contact = empathy_contact_dup_from_tp_contact (tp_contact);
-          empathy_contact_set_persona (contact, FOLKS_PERSONA (persona));
+          if (tp_contact != NULL)
+            {
+              contact = empathy_contact_dup_from_tp_contact (tp_contact);
+              empathy_contact_set_persona (contact, FOLKS_PERSONA (persona));
 
-          g_object_set_data (G_OBJECT (contact), "individual", NULL);
-          g_signal_handlers_disconnect_by_func (contact,
-              (GCallback) individual_store_contact_updated_cb, self);
+              g_object_set_data (G_OBJECT (contact), "individual", NULL);
+              g_signal_handlers_disconnect_by_func (contact,
+                  (GCallback) individual_store_contact_updated_cb, self);
 
-          g_object_unref (contact);
+              g_object_unref (contact);
+            }
         }
 
       g_clear_object (&persona);
@@ -924,16 +934,19 @@ individual_personas_changed_cb (FolksIndividual *individual,
       if (TPF_IS_PERSONA (persona))
         {
           tp_contact = tpf_persona_get_contact (persona);
-          contact = empathy_contact_dup_from_tp_contact (tp_contact);
-          empathy_contact_set_persona (contact, FOLKS_PERSONA (persona));
+          if (tp_contact != NULL)
+            {
+              contact = empathy_contact_dup_from_tp_contact (tp_contact);
+              empathy_contact_set_persona (contact, FOLKS_PERSONA (persona));
 
-          g_object_set_data (G_OBJECT (contact), "individual", individual);
-          g_signal_connect (contact, "notify::capabilities",
-              (GCallback) individual_store_contact_updated_cb, self);
-          g_signal_connect (contact, "notify::client-types",
-              (GCallback) individual_store_contact_updated_cb, self);
+              g_object_set_data (G_OBJECT (contact), "individual", individual);
+              g_signal_connect (contact, "notify::capabilities",
+                  (GCallback) individual_store_contact_updated_cb, self);
+              g_signal_connect (contact, "notify::client-types",
+                  (GCallback) individual_store_contact_updated_cb, self);
 
-          g_object_unref (contact);
+              g_object_unref (contact);
+            }
         }
 
       g_clear_object (&persona);
@@ -1172,6 +1185,8 @@ individual_store_dispose (GObject *object)
     }
 
   g_hash_table_destroy (priv->status_icons);
+  g_hash_table_destroy (priv->folks_individual_cache);
+  g_hash_table_destroy (priv->empathy_group_cache);
   G_OBJECT_CLASS (empathy_individual_store_parent_class)->dispose (object);
 }
 
@@ -1420,25 +1435,28 @@ individual_store_contact_sort (FolksIndividual *individual_a,
 
   contact_a = empathy_contact_dup_from_folks_individual (individual_a);
   contact_b = empathy_contact_dup_from_folks_individual (individual_b);
-  account_a = empathy_contact_get_account (contact_a);
-  account_b = empathy_contact_get_account (contact_b);
+  if (contact_a != NULL && contact_b != NULL)
+    {
+      account_a = empathy_contact_get_account (contact_a);
+      account_b = empathy_contact_get_account (contact_b);
 
-  g_assert (account_a != NULL);
-  g_assert (account_b != NULL);
+      g_assert (account_a != NULL);
+      g_assert (account_b != NULL);
 
-  /* protocol */
-  ret_val = g_strcmp0 (tp_account_get_protocol (account_a),
-      tp_account_get_protocol (account_b));
+      /* protocol */
+      ret_val = g_strcmp0 (tp_account_get_protocol (account_a),
+          tp_account_get_protocol (account_b));
 
-  if (ret_val != 0)
-    goto out;
+      if (ret_val != 0)
+        goto out;
 
-  /* account ID */
-  ret_val = g_strcmp0 (tp_proxy_get_object_path (account_a),
-      tp_proxy_get_object_path (account_b));
+      /* account ID */
+      ret_val = g_strcmp0 (tp_proxy_get_object_path (account_a),
+          tp_proxy_get_object_path (account_b));
 
-  if (ret_val != 0)
-    goto out;
+      if (ret_val != 0)
+        goto out;
+    }
 
   /* identifier */
   ret_val = g_utf8_collate (folks_individual_get_id (individual_a),
@@ -1605,6 +1623,14 @@ individual_store_inhibit_active_cb (EmpathyIndividualStore *self)
   return FALSE;
 }
 
+static void
+g_queue_free_full_row_ref (gpointer data)
+{
+  GQueue *queue = (GQueue *) data;
+  g_queue_foreach (queue, (GFunc) gtk_tree_row_reference_free, NULL);
+  g_queue_free (queue);
+}
+
 static void
 empathy_individual_store_init (EmpathyIndividualStore *self)
 {
@@ -1620,6 +1646,11 @@ empathy_individual_store_init (EmpathyIndividualStore *self)
       (GSourceFunc) individual_store_inhibit_active_cb, self);
   priv->status_icons =
       g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);
+  priv->folks_individual_cache = g_hash_table_new_full (NULL, NULL, NULL,
+      g_queue_free_full_row_ref);
+  priv->empathy_group_cache = g_hash_table_new_full (g_str_hash,
+      g_str_equal, g_free,
+      (GDestroyNotify) gtk_tree_row_reference_free);
   individual_store_setup (self);
 }
 
@@ -1791,6 +1822,10 @@ empathy_individual_store_set_show_groups (EmpathyIndividualStore *self,
       GList *contacts;
 
       gtk_tree_store_clear (GTK_TREE_STORE (self));
+      /* Also clear the cache */
+      g_hash_table_remove_all (priv->folks_individual_cache);
+      g_hash_table_remove_all (priv->empathy_group_cache);
+
       contacts = empathy_individual_manager_get_members (priv->manager);
 
       individual_store_members_changed_cb (priv->manager,
@@ -1992,8 +2027,12 @@ individual_store_get_individual_status_icon_with_icon_name (
   if (show_protocols_here)
     {
       contact = empathy_contact_dup_from_folks_individual (individual);
-      protocol_name = empathy_protocol_name_for_contact (contact);
-      icon_name = g_strdup_printf ("%s-%s", status_icon_name, protocol_name);
+      if (contact != NULL)
+        {
+          protocol_name = empathy_protocol_name_for_contact (contact);
+          icon_name = g_strdup_printf ("%s-%s", status_icon_name,
+              protocol_name);
+        }
     }
   else
     {
